#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the idict project.
#  Please respect the license - more about this in the section (*) below.
#
#  idict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  idict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with idict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.
#
import operator
from functools import reduce
from operator import rshift as aop
from operator import xor as cop
from random import Random
from typing import TypeVar, Union, Callable

from garoupa import ø40
from ldict.core.base import AbstractMutableLazyDict, AbstractLazyDict
from ldict.exception import WrongKeyType

from idict.parameter.ilet import iLet
from idict.parameter.ifunctionspace import iFunctionSpace

VT = TypeVar("VT")


class Idict(AbstractMutableLazyDict):
    """Mutable lazy identified dict for serializable (picklable) pairs str->value

    Usage:

    >>> from idict import idict
    >>> print(idict())
    {
        "id": "0000000000000000000000000000000000000000",
        "ids": {}
    }
    >>> d = idict(x=5, y=3)
    >>> print(d)
    {
        "x": 5,
        "y": 3,
        "id": "Xt_6cc13095bc5b4c671270fbe8ec313568a8b35",
        "ids": {
            "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",
            "y": "XB_1cba4912b6826191bcc15ebde8f1b960282cd"
        }
    }
    >>> d["y"]
    3
    >>> print(idict(x=123123, y=88))
    {
        "x": 123123,
        "y": 88,
        "id": "dR_5b58200b12d6f162541e09c570838ef5a429e",
        "ids": {
            "x": "4W_3331a1c01e3e27831cf08b7bde9b865db7b2e",
            "y": "9X_c8cb257a04eba75c381df365a1e7f7e2dc660"
        }
    }
    >>> print(idict(y=88, x=123123))
    {
        "y": 88,
        "x": 123123,
        "id": "dR_5b58200b12d6f162541e09c570838ef5a429e",
        "ids": {
            "y": "9X_c8cb257a04eba75c381df365a1e7f7e2dc660",
            "x": "4W_3331a1c01e3e27831cf08b7bde9b865db7b2e"
        }
    }
    >>> d = idict(x=123123, y=88)
    >>> d2 = d >> (lambda x: {"z": x**2})
    >>> d2.ids
    {'z': '.JXmafqx65TZ-laengA5qxtk1fUJBi6bgQpYHIM8', 'x': '4W_3331a1c01e3e27831cf08b7bde9b865db7b2e', 'y': '9X_c8cb257a04eba75c381df365a1e7f7e2dc660'}
    >>> d2.hosh == d2.identity * d2.ids["z"] * d2.ids["x"] * d2.ids["y"]
    True
    >>> e = d2 >> (lambda x,y: {"w": x/y})
    >>> print(e)
    {
        "w": "→(x y)",
        "z": "→(x)",
        "x": 123123,
        "y": 88,
        "id": "96PdbhpKgueRWa.LSQWcSSbr.ZMZsuLzkF84sOwe",
        "ids": {
            "w": "1--sDMlN-GuH4FUXhvPWNkyHmTOfTbFo4RK7M5M5",
            "z": ".JXmafqx65TZ-laengA5qxtk1fUJBi6bgQpYHIM8",
            "x": "4W_3331a1c01e3e27831cf08b7bde9b865db7b2e",
            "y": "9X_c8cb257a04eba75c381df365a1e7f7e2dc660"
        }
    }
    >>> a = d >> (lambda x: {"z": x**2}) >> (lambda x, y: {"w": x/y})
    >>> b = d >> (lambda x, y: {"w": x/y}) >> (lambda x: {"z": x**2})
    >>> dic = d.asdict  # Converting to dict
    >>> dic
    {'x': 123123, 'y': 88, 'id': 'dR_5b58200b12d6f162541e09c570838ef5a429e', 'ids': {'x': '4W_3331a1c01e3e27831cf08b7bde9b865db7b2e', 'y': '9X_c8cb257a04eba75c381df365a1e7f7e2dc660'}}
    >>> d2 = idict(dic)  # Reconstructing from a dict
    >>> print(d2)
    {
        "x": 123123,
        "y": 88,
        "id": "dR_5b58200b12d6f162541e09c570838ef5a429e",
        "ids": {
            "x": "4W_3331a1c01e3e27831cf08b7bde9b865db7b2e",
            "y": "9X_c8cb257a04eba75c381df365a1e7f7e2dc660"
        }
    }
    >>> d == d2
    True
    >>> from idict import Ø
    >>> d = Ø >> {"x": "more content"}
    >>> print(d)
    {
        "x": "more content",
        "id": "lU_2bc203cfa982e84748e044ad5f3a86dcf97ff",
        "ids": {
            "x": "lU_2bc203cfa982e84748e044ad5f3a86dcf97ff"
        }
    }
    >>> d = idict() >> {"x": "more content"}
    >>> print(d)
    {
        "x": "more content",
        "id": "lU_2bc203cfa982e84748e044ad5f3a86dcf97ff",
        "ids": {
            "x": "lU_2bc203cfa982e84748e044ad5f3a86dcf97ff"
        }
    }
    >>> e.ids.keys()
    dict_keys(['w', 'z', 'x', 'y'])
    >>> del e["z"]
    >>> print(e)
    {
        "w": "→(x y)",
        "x": 123123,
        "y": 88,
        "id": "GAgXkH4fTORLS1ijp.SQg-6gRa0gTbFo4RK7M5M5",
        "ids": {
            "w": "1--sDMlN-GuH4FUXhvPWNkyHmTOfTbFo4RK7M5M5",
            "x": "4W_3331a1c01e3e27831cf08b7bde9b865db7b2e",
            "y": "9X_c8cb257a04eba75c381df365a1e7f7e2dc660"
        }
    }
    >>> e.hosh == e.identity * e.ids["w"] * e.ids["x"] * e.ids["y"]
    True
    >>> e["x"] = 77
    >>> print(e)
    {
        "w": "→(x y)",
        "x": 77,
        "y": 88,
        "id": "aGMqf9GsQ.SBkKYKE-l21EjPX4YfTbFo4RK7M5M5",
        "ids": {
            "w": "1--sDMlN-GuH4FUXhvPWNkyHmTOfTbFo4RK7M5M5",
            "x": "JF_093a985add7d5e2d319c2662db9ae954648b4",
            "y": "9X_c8cb257a04eba75c381df365a1e7f7e2dc660"
        }
    }
    >>> f = lambda x,y: {"z":x+y}
    >>> d = idict(x=5, y=7)
    >>> d2 = d >> f
    >>> d2.show(colored=False)
    {
        "z": "→(x y)",
        "x": 5,
        "y": 7,
        "id": "M0K6ckhuIW3hnTYCYQ24DmG-H9Fm.mdn2sxVEnRv",
        "ids": {
            "z": "0vOQQX6u2JWqe8DlgbAoZZcKbkIm.mdn2sxVEnRv",
            "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",
            "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8"
        }
    }
    >>> c = {}
    >>> d3 = d2 >> [c]
    >>> d3.show(colored=False)
    {
        "z": "→(^ x y)",
        "x": 5,
        "y": 7,
        "id": "M0K6ckhuIW3hnTYCYQ24DmG-H9Fm.mdn2sxVEnRv",
        "ids": {
            "z": "0vOQQX6u2JWqe8DlgbAoZZcKbkIm.mdn2sxVEnRv",
            "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",
            "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8"
        }
    }
    >>> c
    {}
    >>> d3.z
    12
    >>> c
    {'0vOQQX6u2JWqe8DlgbAoZZcKbkIm.mdn2sxVEnRv': 12, '.T_f0bb8da3062cc75365ae0446044f7b3270977': 5, 'mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8': 7}
    >>> d3.show(colored=False)
    {
        "z": 12,
        "x": 5,
        "y": 7,
        "id": "M0K6ckhuIW3hnTYCYQ24DmG-H9Fm.mdn2sxVEnRv",
        "ids": {
            "z": "0vOQQX6u2JWqe8DlgbAoZZcKbkIm.mdn2sxVEnRv",
            "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",
            "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8"
        }
    }
    >>> c = {}
    >>> from idict import setup
    >>> setup(cache=c)
    >>> d3 = d >> f ^ Ø
    >>> d3.show(colored=False)
    {
        "z": "→(^ x y)",
        "x": 5,
        "y": 7,
        "id": "M0K6ckhuIW3hnTYCYQ24DmG-H9Fm.mdn2sxVEnRv",
        "ids": {
            "z": "0vOQQX6u2JWqe8DlgbAoZZcKbkIm.mdn2sxVEnRv",
            "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",
            "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8"
        }
    }
    >>> c
    {}
    >>> d3.z
    12
    >>> c
    {'0vOQQX6u2JWqe8DlgbAoZZcKbkIm.mdn2sxVEnRv': 12, '.T_f0bb8da3062cc75365ae0446044f7b3270977': 5, 'mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8': 7}
    >>> d3.show(colored=False)
    {
        "z": 12,
        "x": 5,
        "y": 7,
        "id": "M0K6ckhuIW3hnTYCYQ24DmG-H9Fm.mdn2sxVEnRv",
        "ids": {
            "z": "0vOQQX6u2JWqe8DlgbAoZZcKbkIm.mdn2sxVEnRv",
            "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",
            "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8"
        }
    }
    """

    # noinspection PyMissingConstructor
    def __init__(self, /, _dictionary=None, id=None, ids=None, rnd=None, identity=ø40, _cloned=None, **kwargs):
        self.identity = identity
        from idict.core.frozenidentifieddict import FrozenIdentifiedDict
        self.frozen: FrozenIdentifiedDict = FrozenIdentifiedDict(_dictionary, id, ids, rnd, identity, _cloned, **kwargs)

    @property
    def hosh(self):
        return self.frozen.hosh

    @property
    def blobs(self):
        return self.frozen.blobs

    @property
    def hashes(self):
        return self.frozen.hashes

    @property
    def hoshes(self):
        return self.frozen.hoshes

    def __delitem__(self, key):
        if not isinstance(key, str):
            raise WrongKeyType(f"Key must be string, not {type(key)}.", key)
        data, blobs, hashes, hoshes = self.data.copy(), self.blobs.copy(), self.hashes.copy(), self.hoshes.copy()
        del data[key]
        for coll in [blobs, hashes, hoshes]:
            if key in coll:
                del coll[key]
        hosh = reduce(operator.mul, [self.identity] + list(hoshes.values()))
        self.frozen = self.frozen.clone(data, _cloned=dict(blobs=blobs, hashes=hashes, hoshes=hoshes, hosh=hosh))

    def clone(self, data=None, rnd=None, _cloned=None):
        cloned_internals = _cloned or dict(blobs=self.blobs, hashes=self.hashes, hoshes=self.hoshes, hosh=self.hosh)
        return self.__class__(data or self.data, rnd=rnd or self.rnd, identity=self.identity, _cloned=cloned_internals)

    def show(self, colored=True, width=None):
        self.frozen.show(colored, width)

    def __rrshift__(self, left: Union[Random, dict, Callable, iFunctionSpace]):
        """
        >>> print({"x": 5} >> Idict(y=2))
        {
            "x": 5,
            "y": 2,
            "id": "o8_4c07d34b8963338a275e43bfcac9c37f125cc",
            "ids": {
                "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",
                "y": "pg_7d1eecc7838558a4c1bf9584d68a487791c45"
            }
        }
        >>> print((lambda x: {"y": 5*x}) >> Idict(y=2))
        «λ × {
            "y": 2,
            "id": "pg_7d1eecc7838558a4c1bf9584d68a487791c45",
            "ids": {
                "y": "pg_7d1eecc7838558a4c1bf9584d68a487791c45"
            }
        }»
        """
        if isinstance(left, list) or callable(left):
            return iFunctionSpace(left, aop, self)
        clone = self.__class__(identity=self.identity)
        clone.frozen = left >> self.frozen
        return clone

    def __rshift__(self, other: Union[dict, AbstractLazyDict, Callable, iLet, iFunctionSpace, Random]):
        """
        >>> d = Idict(x=2) >> (lambda x: {"y": 2 * x})
        >>> d.ids
        {'y': 'zJmLy1B8VQU8.Kji0iqU0zIrDWpWqcXxhrGWdepm', 'x': 'og_0f0d4c16437fb2a4c1bff594d68a486791c45'}
        """
        clone = self.__class__(identity=self.identity)
        clone.frozen = self.frozen >> other
        return clone

    def __rxor__(self, left: Union[Random, dict, Callable, iFunctionSpace]):
        if isinstance(left, list) or callable(left):
            return iFunctionSpace(left, cop, self)
        clone = self.__class__(identity=self.identity)
        clone.frozen = left ^ self.frozen
        return clone

    def __xor__(self, other: Union[dict, AbstractLazyDict, Callable, iLet, iFunctionSpace, Random]):
        clone = self.__class__(identity=self.identity)
        clone.frozen = self.frozen ^ other
        return clone



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the idict project.
#  Please respect the license - more about this in the section (*) below.
#
#  idict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  idict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with idict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.
from functools import cached_property
from json import dumps
from operator import rshift as aop
from operator import xor as cop
from random import Random
from typing import Union, Callable

from ldict.core.base import AbstractLazyDict
from ldict.customjson import CustomJSONEncoder
from ldict.parameter.abslet import AbstractLet


class iLet(AbstractLet):
    """
    Set values or sampling intervals for parameterized functions

    >>> from idict import idict, let
    >>> f = lambda x,y, a=[-1,-0.9,-0.8,...,1]: {"z": a*x + y}
    >>> f_a = let(f, a=0)
    >>> f_a
    λ{'a': 0}
    >>> d = idict(x=5,y=7)
     >>> d2 = d >> f_a
    >>> print(d2)
    {
        "z": "→(a x y)",
        "x": 5,
        "y": 7,
        "id": "tsKSa7giKVTZC-s0qQVQy3nO923A59f327Jd.05S",
        "ids": {
            "z": "Cv907SXx1qXi0b7LJar9VGVxFc6A59f327Jd.05S",
            "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",
            "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8"
        }
    }
    >>> d2.evaluate()
    >>> print(d2)
    {
        "z": 7,
        "x": 5,
        "y": 7,
        "id": "tsKSa7giKVTZC-s0qQVQy3nO923A59f327Jd.05S",
        "ids": {
            "z": "Cv907SXx1qXi0b7LJar9VGVxFc6A59f327Jd.05S",
            "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",
            "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8"
        }
    }
    >>> from random import Random
    >>> d2 = d >> Random(0) >> let(f, a=[8,9])
    >>> print(d2)
    {
        "z": "→(a x y)",
        "x": 5,
        "y": 7,
        "id": "qMwvsA4Ll3c7CLVlafJo6o3yP5dEhLZPNh9QkVgv",
        "ids": {
            "z": "Fir2J-MamdlPkZz4uBeJs.BhjggEhLZPNh9QkVgv",
            "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",
            "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8"
        }
    }
    >>> d2.evaluate()
    >>> print(d2)
    {
        "z": 52,
        "x": 5,
        "y": 7,
        "id": "qMwvsA4Ll3c7CLVlafJo6o3yP5dEhLZPNh9QkVgv",
        "ids": {
            "z": "Fir2J-MamdlPkZz4uBeJs.BhjggEhLZPNh9QkVgv",
            "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",
            "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8"
        }
    }
    >>> let(f, a=5) >> {"x": 5, "y": 7}
    «λ{'a': 5} × {'x': 5, 'y': 7}»
    >>> print(idict({"x": 5, "y": 7}) >> let(f, a=5))
    {
        "z": "→(a x y)",
        "x": 5,
        "y": 7,
        "id": "rrC7MbIwMVR4rAPcSEgs3kSbJnJuCMS1F1k3.VfQ",
        "ids": {
            "z": "XLoMEFGNqF52tTtX9.NMpXoXcyMuCMS1F1k3.VfQ",
            "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",
            "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8"
        }
    }
    >>> let(f, a=5) >> idict({"x": 5, "y": 7})
    «λ{'a': 5} × {
        "x": 5,
        "y": 7,
        "id": "mP_2d615fd34f97ac906e162c6fc6aedadc4d140",
        "ids": {
            "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",
            "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8"
        }
    }»
    >>> let(f, a=5) >> ["mycache"]
    «λ{'a': 5} × ^»
    >>> from idict.parameter.ifunctionspace import iFunctionSpace
    >>> let(f, a=5) >> iFunctionSpace()
    «λ{'a': 5}»
    >>> iFunctionSpace() >> let(f, a=5)
    «λ{'a': 5}»
    >>> (lambda x: {"z": x*8}) >> let(f, a=5)
    «λ × λ{'a': 5}»
    >>> d = {"x":3, "y": 8} >> let(f, a=5)
    >>> print(d)
    {
        "z": "→(a x y)",
        "x": 3,
        "y": 8,
        "id": "cKaaEdvxmlKTqsyOsb8eAJMl4aCuCMS1F1k3.VfQ",
        "ids": {
            "z": "7WQXcDiKAKIjnfIwETM9UknKD9CuCMS1F1k3.VfQ",
            "x": "WB_e55a47230d67db81bcc1aecde8f1b950282cd",
            "y": "6q_07bbf68ac6eb0f9e2da3bda1665567bc21bde"
        }
    }
    >>> print(d.z)
    23
    >>> d >>= Random(0) >> let(f, a=[1,2,3]) >> let(f, a=[9,8,7])
    >>> print(d)
    {
        "z": "→(a x y)",
        "x": 3,
        "y": 8,
        "id": "5Q7U4sm-DkzdUwGKY.9IrQRTpFLMIhJOSvWYkxt-",
        "ids": {
            "z": "oOhgGymI3OqJG1As8IODMrsgZELMIhJOSvWYkxt-",
            "x": "WB_e55a47230d67db81bcc1aecde8f1b950282cd",
            "y": "6q_07bbf68ac6eb0f9e2da3bda1665567bc21bde"
        }
    }
    >>> print(d.z)
    32
    """

    def __init__(self, f, **kwargs):
        self.f = f
        self.config = {k: kwargs[k] for k in sorted(kwargs.keys())}
        self.asdict = self.config

    @cached_property
    def bytes(self):
        return dumps(self.config, sort_keys=True, cls=CustomJSONEncoder).encode()

    def __repr__(self):
        return "λ" + str(self.config)

    def __rrshift__(self, left: Union[dict, list, Random, Callable, 'iLet']):
        """
        >>> from idict.parameter.ilet import iLet
        >>> {"x":5} >> iLet(lambda x:{"x": x**2}, x=5)
        {
            "x": "→(x)",
            "id": "LEMLzqy0ijWZcJ8w37f8QE2tj-7QOh11-DoSoW4j",
            "ids": "LEMLzqy0ijWZcJ8w37f8QE2tj-7QOh11-DoSoW4j"
        }
        >>> [1] >> iLet(lambda x:{"x": x**2}, x=5)
        «^ × λ{'x': 5}»
        """
        if isinstance(left, dict) and not isinstance(left, AbstractLazyDict):
            from idict.core.idict_ import Idict
            return Idict(left) >> self
        if isinstance(left, (list, Random, Callable)):
            from idict.parameter.ifunctionspace import iFunctionSpace
            return iFunctionSpace(left, aop, self)
        return NotImplemented  # pragma: no cover

    def __rshift__(self, other: Union[dict, list, Random, Callable, 'iLet', AbstractLazyDict]):
        """
        >>> iLet(lambda x:{"x": x**2}, x=5) >> [1]
        «λ{'x': 5} × ^»
        """

        if isinstance(other, (dict, list, Random, Callable, iLet)):
            from idict.parameter.ifunctionspace import iFunctionSpace
            return iFunctionSpace(self, aop, other)
        return NotImplemented  # pragma: no cover

    def __rxor__(self, left: Union[dict, list, Random, Callable, 'iLet']):
        if isinstance(left, (dict, list, Random, Callable)) and not isinstance(left, AbstractLazyDict):
            from idict.parameter.ifunctionspace import iFunctionSpace
            return iFunctionSpace(left, cop, self)
        return NotImplemented  # pragma: no cover

    def __xor__(self, other: Union[dict, list, Random, Callable, 'iLet', AbstractLazyDict]):
        if isinstance(other, (dict, list, Random, Callable, iLet)):
            from idict.parameter.ifunctionspace import iFunctionSpace
            return iFunctionSpace(self, cop, other)
        return NotImplemented  # pragma: no cover



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the idict project.
#  Please respect the license - more about this in the section (*) below.
#
#  idict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  idict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with idict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.
from typing import TypeVar

from idict.core.compression import unpack, pack
from idict.persistence.cache import Cache
from idict.persistence.raw.sqladict import sqladict, check, Content

VT = TypeVar("VT")


class SQLA(Cache):  # pragma:  cover
    """Save to/retrieve from SQLAlchemy.

    Based on built-in module shelve. Open and close at every transaction.
    To keep open, please use shelve context manager itself.

    >>> d = SQLA("sqlite+pysqlite:////tmp/sqla-doctest.db")
    >>> d["x"] = 5
    >>> d["x"]
    5
    >>> for k,v in d.items():
    ...     print(k, v)
    x 5
    >>> "x" in d
    True
    >>> len(d)
    1
    >>> del d["x"]
    >>> "x" in d
    False
    >>> d
    SQLA→<class 'idict.persistence.raw.sqladict.SQLAdict'>
    """

    def __init__(self, url="sqlite+pysqlite:///:memory:", debug=False):
        super().__init__(lambda: sqladict(url, debug))

    def __setitem__(self, key: str, value):
        check(key)
        super().__setitem__(key, pack(value))

    def __getitem__(self, key):
        check(key)
        return unpack(super().__getitem__(key))

    def __iter__(self):
        with self.decorator() as db:
            return (c.id for c in db.session.query(Content).all())

    def __len__(self):
        with self.decorator() as db:
            return db.session.query(Content).count()

    def copy(self):
        raise NotImplementedError


# TODO: fazer decorator sqla



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the idict project.
#  Please respect the license - more about this in the section (*) below.
#
#  idict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  idict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with idict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.

from abc import ABC, abstractmethod
from typing import TypeVar

VT = TypeVar("VT")


class Cache(ABC):  # pragma: no cover
    def __init__(self, decorator):
        self.decorator = decorator

    def __contains__(self, item):
        with self.decorator() as db:
            return item in db

    def __setitem__(self, key, value):
        with self.decorator() as db:
            db[key] = value

    def __getitem__(self, key):
        with self.decorator() as db:
            return db[key]

    def __delitem__(self, key):
        with self.decorator() as db:
            del db[key]

    @abstractmethod
    def __iter__(self):
        raise NotImplementedError

    def __repr__(self):
        with self.decorator() as db:
            return self.__class__.__name__ + "→" + str(type(db))

    def copy(self):
        with self.decorator() as db:
            dic = dict(db)
            return dic

    def keys(self):
        return iter(self)

    def items(self):
        for k in self.keys():
            yield k, self[k]



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the idict project.
#  Please respect the license - more about this in the section (*) below.
#
#  idict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  idict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with idict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.
#
import operator
from functools import reduce
from operator import rshift as aop
from operator import xor as cop
from random import Random
from typing import TypeVar, Union, Callable

from garoupa import ø40, Hosh
from ldict.core.base import AbstractLazyDict, AbstractMutableLazyDict
from ldict.customjson import CustomJSONEncoder
from ldict.frozenlazydict import FrozenLazyDict

from idict.config import GLOBAL
from idict.core.appearance import decolorize, idict2txt
from idict.core.identification import key2id, blobs_hashes_hoshes
from idict.parameter.ifunctionspace import iFunctionSpace, reduce3
from idict.parameter.ilet import iLet
from idict.persistence.cached import cached

VT = TypeVar("VT")


class FrozenIdentifiedDict(AbstractLazyDict):
    """Immutable lazy universally identified dict for serializable (picklable) pairs str->value

    Usage:

    >>> idict = FrozenIdentifiedDict
    >>> print(idict())
    {
        "id": "0000000000000000000000000000000000000000",
        "ids": {}
    }
    >>> d = idict(x=5, y=3)
    >>> print(d)
    {
        "x": 5,
        "y": 3,
        "id": "Xt_6cc13095bc5b4c671270fbe8ec313568a8b35",
        "ids": {
            "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",
            "y": "XB_1cba4912b6826191bcc15ebde8f1b960282cd"
        }
    }
    >>> d["y"]
    3
    >>> print(idict(x=123123, y=88))
    {
        "x": 123123,
        "y": 88,
        "id": "dR_5b58200b12d6f162541e09c570838ef5a429e",
        "ids": {
            "x": "4W_3331a1c01e3e27831cf08b7bde9b865db7b2e",
            "y": "9X_c8cb257a04eba75c381df365a1e7f7e2dc660"
        }
    }
    >>> print(idict(y=88, x=123123))
    {
        "y": 88,
        "x": 123123,
        "id": "dR_5b58200b12d6f162541e09c570838ef5a429e",
        "ids": {
            "y": "9X_c8cb257a04eba75c381df365a1e7f7e2dc660",
            "x": "4W_3331a1c01e3e27831cf08b7bde9b865db7b2e"
        }
    }
    >>> d = idict(x=123123, y=88)
    >>> d2 = d >> (lambda x: {"z": x**2})
    >>> d2.hosh == d2.identity * d2.ids["z"] * d2.ids["x"] * d2.ids["y"]
    True
    >>> e = d2 >> (lambda x,y: {"w": x/y})
    >>> print(e)
    {
        "w": "→(x y)",
        "z": "→(x)",
        "x": 123123,
        "y": 88,
        "id": "96PdbhpKgueRWa.LSQWcSSbr.ZMZsuLzkF84sOwe",
        "ids": {
            "w": "1--sDMlN-GuH4FUXhvPWNkyHmTOfTbFo4RK7M5M5",
            "z": ".JXmafqx65TZ-laengA5qxtk1fUJBi6bgQpYHIM8",
            "x": "4W_3331a1c01e3e27831cf08b7bde9b865db7b2e",
            "y": "9X_c8cb257a04eba75c381df365a1e7f7e2dc660"
        }
    }
    >>> a = d >> (lambda x: {"z": x**2}) >> (lambda x, y: {"w": x/y})
    >>> b = d >> (lambda x, y: {"w": x/y}) >> (lambda x: {"z": x**2})
    >>> dic = d.asdict  # Converting to dict
    >>> dic
    {'x': 123123, 'y': 88, 'id': 'dR_5b58200b12d6f162541e09c570838ef5a429e', 'ids': {'x': '4W_3331a1c01e3e27831cf08b7bde9b865db7b2e', 'y': '9X_c8cb257a04eba75c381df365a1e7f7e2dc660'}}
    >>> d2 = idict(dic)  # Reconstructing from a dict
    >>> print(d2)
    {
        "x": 123123,
        "y": 88,
        "id": "dR_5b58200b12d6f162541e09c570838ef5a429e",
        "ids": {
            "x": "4W_3331a1c01e3e27831cf08b7bde9b865db7b2e",
            "y": "9X_c8cb257a04eba75c381df365a1e7f7e2dc660"
        }
    }
    >>> d == d2
    True
    >>> from idict import Ø, setup
    >>> d = idict() >> {"x": "more content"}
    >>> print(d)
    {
        "x": "more content",
        "id": "lU_2bc203cfa982e84748e044ad5f3a86dcf97ff",
        "ids": {
            "x": "lU_2bc203cfa982e84748e044ad5f3a86dcf97ff"
        }
    }
    >>> f = lambda x,y: {"z":x+y}
    >>> d = idict(x=5, y=7)
    >>> d2 = d >> f
    >>> d2.show(colored=False)
    {
        "z": "→(x y)",
        "x": 5,
        "y": 7,
        "id": "M0K6ckhuIW3hnTYCYQ24DmG-H9Fm.mdn2sxVEnRv",
        "ids": {
            "z": "0vOQQX6u2JWqe8DlgbAoZZcKbkIm.mdn2sxVEnRv",
            "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",
            "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8"
        }
    }
    >>> c = {}
    >>> d3 = d2 >> [c]
    >>> d3.show(colored=False)
    {
        "z": "→(^ x y)",
        "x": 5,
        "y": 7,
        "id": "M0K6ckhuIW3hnTYCYQ24DmG-H9Fm.mdn2sxVEnRv",
        "ids": {
            "z": "0vOQQX6u2JWqe8DlgbAoZZcKbkIm.mdn2sxVEnRv",
            "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",
            "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8"
        }
    }
    >>> c
    {}
    >>> d3.z
    12
    >>> c
    {'0vOQQX6u2JWqe8DlgbAoZZcKbkIm.mdn2sxVEnRv': 12, '.T_f0bb8da3062cc75365ae0446044f7b3270977': 5, 'mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8': 7}
    >>> d3.show(colored=False)
    {
        "z": 12,
        "x": 5,
        "y": 7,
        "id": "M0K6ckhuIW3hnTYCYQ24DmG-H9Fm.mdn2sxVEnRv",
        "ids": {
            "z": "0vOQQX6u2JWqe8DlgbAoZZcKbkIm.mdn2sxVEnRv",
            "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",
            "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8"
        }
    }
    >>> c = {}
    >>> setup(cache=c)
    >>> d3 = d >> f ^ Ø
    >>> d3.show(colored=False)
    {
        "z": "→(^ x y)",
        "x": 5,
        "y": 7,
        "id": "M0K6ckhuIW3hnTYCYQ24DmG-H9Fm.mdn2sxVEnRv",
        "ids": {
            "z": "0vOQQX6u2JWqe8DlgbAoZZcKbkIm.mdn2sxVEnRv",
            "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",
            "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8"
        }
    }
    >>> c
    {}
    >>> d3.z
    12
    >>> c
    {'0vOQQX6u2JWqe8DlgbAoZZcKbkIm.mdn2sxVEnRv': 12, '.T_f0bb8da3062cc75365ae0446044f7b3270977': 5, 'mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8': 7}
    >>> d3.show(colored=False)
    {
        "z": 12,
        "x": 5,
        "y": 7,
        "id": "M0K6ckhuIW3hnTYCYQ24DmG-H9Fm.mdn2sxVEnRv",
        "ids": {
            "z": "0vOQQX6u2JWqe8DlgbAoZZcKbkIm.mdn2sxVEnRv",
            "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",
            "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8"
        }
    }
    """
    hosh: Hosh

    # noinspection PyMissingConstructor
    def __init__(self, /, _dictionary=None, id=None, ids=None, rnd=None, identity=ø40, _cloned=None, **kwargs):
        self.rnd = rnd
        self.identity = identity
        data = _dictionary or {}
        data.update(kwargs)

        # Freeze mutable *dicts.
        for k, v in data.items():
            if isinstance(v, AbstractMutableLazyDict):
                data[k] = v.frozen

        if _cloned:
            self.blobs = _cloned["blobs"]
            self.hashes = _cloned["hashes"]
            self.hoshes = _cloned["hoshes"]
            self.hosh = _cloned["hosh"]
        else:
            if "id" in data:
                if id:  # pragma: no cover
                    raise Exception(f"Conflicting 'id' values: {id} and {data['id']}")
                id = data.pop("id")
            if "ids" in data:
                if ids:  # pragma: no cover
                    raise Exception(f"Conflicting 'ids' values: {ids} and {data['ids']}")
                ids = data.pop("ids")

            if id:
                if ids is None:  # pragma: no cover
                    raise Exception(f"'id' {id} given, but 'ids' is missing.")
                self.blobs = {}
                self.hashes = {}
                self.hoshes = {k: identity * v for k, v in ids.items()}
            else:
                self.blobs, self.hashes, self.hoshes = blobs_hashes_hoshes(data, identity, ids or {}).values()
            self.hosh = reduce(operator.mul, [identity] + list(self.hoshes.values()))

        if id is None:
            id = self.hosh.id
            try:
                ids = {k: v.id for k, v in self.hoshes.items()}
            except:
                print(self.hoshes)
                raise Exception()

        # Store as an immutable lazy dict.
        self.frozen = FrozenLazyDict(data, id=id, ids=ids, rnd=rnd)
        self.data = self.frozen.data
        self.id = self.hosh.id

    def __getitem__(self, item):
        return self.frozen[item]

    def __setitem__(self, key: str, value):
        self.frozen[key] = value

    def __delitem__(self, key):
        del self.frozen[key]

    def __getattr__(self, item):
        if item == "all":
            return self.__repr__(all=True)
        if item == "asdict":
            return self.frozen.asdict
        return getattr(self.frozen, item)

    def __repr__(self):
        return repr(self.frozen)

    __str__ = __repr__

    def evaluate(self):
        """
        >>> from idict.core.frozenidentifieddict import FrozenIdentifiedDict as idict
        >>> f = lambda x: {"y": x+2}
        >>> d = idict(x=3)
        >>> a = d >> f
        >>> print(a)
        {
            "y": "→(x)",
            "x": 3,
            "id": "tFkvrmyHlXSnstVFIFktJjD7K91yW4AU0sYuSnwe",
            "ids": {
                "y": "BZz1P5xA5r0gfAqOtHySEb.m0HTxW4AU0sYuSnwe",
                "x": "WB_e55a47230d67db81bcc1aecde8f1b950282cd"
            }
        }
        >>> a.evaluate()
        >>> print(a)
        {
            "y": 5,
            "x": 3,
            "id": "tFkvrmyHlXSnstVFIFktJjD7K91yW4AU0sYuSnwe",
            "ids": {
                "y": "BZz1P5xA5r0gfAqOtHySEb.m0HTxW4AU0sYuSnwe",
                "x": "WB_e55a47230d67db81bcc1aecde8f1b950282cd"
            }
        }
        """
        self.frozen.evaluate()

    @property
    def asdict(self):
        """
        >>> from idict.core.frozenidentifieddict import FrozenIdentifiedDict as idict
        >>> d = idict(x=3, y=5)
        >>> d.id
        'Xt_a63010fa2b5b4c671270fbe8ec313568a8b35'
        >>> e = idict(x=7, y=8, d=d)
        >>> e.asdict
        {'x': 7, 'y': 8, 'd': {'x': 3, 'y': 5, 'id': 'Xt_a63010fa2b5b4c671270fbe8ec313568a8b35', 'ids': {'x': 'WB_e55a47230d67db81bcc1aecde8f1b950282cd', 'y': '0U_e2a86ff72e226d5365aea336044f7b4270977'}}, 'id': 'AN_650bae25143e28c5489bfbc806f5fb55c6fdc', 'ids': {'x': 'lX_9e55978592eeb1caf8778e34d26f5fd4cc8c8', 'y': '6q_07bbf68ac6eb0f9e2da3bda1665567bc21bde', 'd': '8s_1ccd1655bae1d9e91270e5eddc31351eb8b35'}}
        >>> d.hosh ** key2id("d", d.identity.digits) == e.hoshes["d"]
        True
        """
        return self.frozen.asdict

    def clone(self, data=None, rnd=None, _cloned=None):
        cloned_internals = _cloned or dict(blobs=self.blobs, hashes=self.hashes, hoshes=self.hoshes, hosh=self.hosh)
        return FrozenIdentifiedDict(
            data or self.data, rnd=rnd or self.rnd, identity=self.identity, _cloned=cloned_internals
        )

    def __hash__(self):
        return hash(self.hosh)

    def show(self, colored=True, width=None):
        r"""
        >>> idict = FrozenIdentifiedDict
        >>> idict(x=134124, y= 56).show(colored=False)
        {
            "x": 134124,
            "y": 56,
            "id": "dq_d85091ef315b9ce0d5eb1a5aabb6e6434a97f",
            "ids": {
                "x": "gZ_37ee5e71c9cd4c9bde421cdb917e5c56f7ebe",
                "y": "Zs_c473399e77e6c2d2f69914891a488a3732bb0"
            }
        }
        """
        CustomJSONEncoder.width = width
        return print(self.all if colored else decolorize(self.all))

    def __repr__(self, all=False):
        return idict2txt(self, all)

    @property
    def all(self):
        r"""
        Usage:

        >>> from idict.core.frozenidentifieddict import FrozenIdentifiedDict as idict
        >>> from idict.core.appearance import decolorize
        >>> out = idict(x=134124, y= 56).all
        >>> decolorize(out)
        '{\n    "x": 134124,\n    "y": 56,\n    "id": "dq_d85091ef315b9ce0d5eb1a5aabb6e6434a97f",\n    "ids": {\n        "x": "gZ_37ee5e71c9cd4c9bde421cdb917e5c56f7ebe",\n        "y": "Zs_c473399e77e6c2d2f69914891a488a3732bb0"\n    }\n}'
        """
        return self.__repr__(all=True)

    def __eq__(self, other):
        """
        >>> from idict import idict
        >>> idict(x=3) == {"x": 3}
        True
        >>> idict(x=3) == {"x": 3, "id": idict(x=3).id}
        True
        >>> idict(x=3) == idict(x=3)
        True
        >>> idict(x=3) != {"x": 4}
        True
        >>> idict(x=3) != idict(x=4)
        True
        >>> idict(x=3) != {"y": 3}
        True
        >>> idict(x=3) != {"x": 3, "id": (~idict(x=3).hosh).id}
        True
        >>> idict(x=3) != idict(y=3)
        True
        """
        if isinstance(other, dict):
            if "id" in other:
                return self.id == other["id"]
            if list(self.keys())[:-2] != list(other.keys()):
                return False
        from idict.core.idict_ import Idict
        if isinstance(other, (FrozenIdentifiedDict, Idict)):
            return self.hosh == other.hosh
        if isinstance(other, AbstractLazyDict):
            if self.keys() != other.keys():
                return False
            other.evaluate()
            return self.data == other.data
        if isinstance(other, dict):
            data = self.data.copy()
            del data["id"]
            del data["ids"]
            return data == other
        raise TypeError(f"Cannot compare {type(self)} and {type(other)}")  # pragma: no cover

    def __rrshift__(self, left: Union[Random, dict, Callable, iFunctionSpace]):
        if isinstance(left, dict) and not isinstance(left, AbstractLazyDict):
            return FrozenIdentifiedDict(left) >> self
        if isinstance(left, list) or callable(left):
            return iFunctionSpace(left, aop, self)
        if isinstance(left, Random):
            return self.clone(rnd=left)
        return NotImplemented

    def __rshift__(self, other: Union[dict, AbstractLazyDict, Callable, iLet, iFunctionSpace, Random]):
        from idict.core.rshift import application, ihandle_dict
        if isinstance(other, dict):
            return ihandle_dict(self, other)
        if isinstance(other, Random):
            return self.clone(rnd=other)
        if isinstance(other, iLet):
            return application(self, other, other.f, other.bytes)
        if callable(other):
            return application(self, other, other, self.identity)
        if isinstance(other, list):
            d = self
            for cache in other:
                d = cached(d, cache)
            return d
        if isinstance(other, iFunctionSpace):
            return reduce3(lambda a, op, b: op(a, b), (self, aop) + other.functions)
        return NotImplemented

    def __rxor__(self, left: Union[Random, dict, Callable, iFunctionSpace]):
        if (isinstance(left, (dict, list, Random)) or callable(left)) and not isinstance(left, AbstractLazyDict):
            return iFunctionSpace(left, cop, self)
        return NotImplemented

    def __xor__(self, other: Union[dict, AbstractLazyDict, Callable, iLet, iFunctionSpace, Random]):
        if callable(other) or isinstance(other, (dict, Random, list, iLet)):
            return cached(self, GLOBAL["cache"]) >> other
        if isinstance(other, iFunctionSpace):
            return reduce3(lambda a, op, b: op(a, b), (self, cop) + other.functions)
        return NotImplemented



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the idict project.
#  Please respect the license - more about this in the section (*) below.
#
#  idict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  idict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with idict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.
from ldict.lazyval import LazyVal


def cached(d, cache):
    def closure(outputf, id, ids, data, output_fields):

        def func(**kwargs):
            # Try loading.
            if id in cache:
                return cache[id]

            # Process and save (all fields, to avoid a parcial ldict being stored).
            result = k = None
            for k, fid in ids.items():
                # TODO: all lazies are evaluated, but show() still shows deps as lazy.
                #    Fortunately the dep is evaluated only once.
                if isinstance(data[k], LazyVal):
                    data[k] = data[k](**kwargs)
                cache[fid] = data[k]
                if k == outputf:
                    result = data[k]
            if result is None:
                if k is None:
                    raise Exception(f"No ids")
                raise Exception(f"{k=} not in output fields: {output_fields}. ids: {ids.items()}")

            # Return requested value.
            return result

        return func

    data = d.data.copy()
    lazies = False
    output_fields = []
    for field, v in list(data.items())[:-2]:
        if isinstance(v, LazyVal):
            output_fields.append(field)
            lazies = True
            id = d.hashes[field].id if field in d.hashes else d.hoshes[field].id
            deps = {"^": None}
            deps.update(v.deps)
            lazy = LazyVal(field, closure(field, id, d.ids, d.data, output_fields), deps, None)
            data[field] = lazy

    # Eager saving when there are no lazies.
    if not lazies:
        for k, v in list(data.items())[:-2]:
            if k not in cache:
                cache[k] = data[k]

    return d.clone(data)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['idict',
 'idict.core',
 'idict.parameter',
 'idict.persistence',
 'idict.persistence.raw']

package_data = \
{'': ['*']}

install_requires = \
['ldict>=3.211031.1,<4.0.0',
 'lz4>=3.1.3,<4.0.0',
 'orjson>=3.5.0,<4.0.0',
 'pdoc3>=0.10.0,<0.11.0']

extras_require = \
{'full': ['pandas>=1.3.3,<2.0.0']}

setup_kwargs = {
    'name': 'idict',
    'version': '0.211117.1',
    'description': 'Lazy dict with universally unique identifier for values',
    'long_description': '![test](https://github.com/davips/idict/workflows/test/badge.svg)\n[![codecov](https://codecov.io/gh/davips/idict/branch/main/graph/badge.svg)](https://codecov.io/gh/davips/idict)\n<a href="https://pypi.org/project/idict">\n<img src="https://img.shields.io/pypi/v/idict.svg?label=release&color=blue&style=flat-square" alt="pypi">\n</a>\n![Python version](https://img.shields.io/badge/python-3.8%20%7C%203.9-blue.svg)\n[![license: GPL v3](https://img.shields.io/badge/License-GPLv3-blue.svg)](https://www.gnu.org/licenses/gpl-3.0)\n\n<!--- [![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.5501845.svg)](https://doi.org/10.5281/zenodo.5501845) --->\n[![arXiv](https://img.shields.io/badge/arXiv-2109.06028-b31b1b.svg?style=flat-square)](https://arxiv.org/abs/2109.06028)\n[![API documentation](https://img.shields.io/badge/doc-API%20%28auto%29-a0a0a0.svg)](https://davips.github.io/idict)\n\n# idict\n\nA lazy `dict` with universally unique deterministic identifiers.\n\n[Latest release](https://pypi.org/project/idict) |\n[Current code](https://github.com/davips/idict) |\n[API documentation](https://davips.github.io/idict)\n\n### See also\n\n* identification package used by `idict`: [GaROUPa](https://pypi.org/project/garoupa)\n* only laziness, i.e., without the identification part: [ldict](https://pypi.org/project/ldict)\n\n## Overview\n\nAn `idict` is an identified `dict` with `str` keys.\nWe consider that every value is generated by a process, starting from an `empty` `idict`. The process is a sequence of\ntransformation steps done through the operator `>>`, which symbolizes the ordering of the steps.\nThere are two types of steps:\n\n* **value insertion** - represented by dict-like objects\n* **function application** - represented by ordinary Python functions\n\nFunctions, `idict`s, and values have a deterministic UUID\n(called _hosh_ - **o**perable **h**a**sh**). \nIdentifiers (hoshes) for `idict`s and values are predictable through the\nmagic available [here](https://pypi.org/project/garoupa).\nAn `idict` is completely defined by its key-value pairs so that\nit can be converted from/to a built-in `dict`.\n\nCreating an `idict` is not different from creating an ordinary `dict`. Optionally it can be created through the `>>` operator\nused after `empty` or `Ø` (usually AltGr+Shift+o in most keyboards).\nThe resulting `idict` always contains two extra entries `id` and `ids`:\n![img.png](https://raw.githubusercontent.com/davips/idict/main/examples/img.png)\n\nFunction application is done in the same way. The parameter names define the input fields, while the keys in the\nreturned `dict` define the output fields:\n![img_1.png](https://raw.githubusercontent.com/davips/idict/main/examples/img_1.png)\n\nAfter evaluated, the value will not be calculated again:\n![img_2.png](https://raw.githubusercontent.com/davips/idict/main/examples/img_2.png)\n\nFunctions can accept parameters:\n![img_3.png](https://raw.githubusercontent.com/davips/idict/main/examples/img_3.png)\n\n\n## Installation\n### ...as a standalone lib\n```bash\n# Set up a virtualenv. \npython3 -m venv venv\nsource venv/bin/activate\n\n# Install from PyPI...\npip install --upgrade pip\npip install -U idict\npip install -U idict[full]  # use the flag \'full\' for extra functionality (recommended)\n\n# ...or, install from updated source code.\npip install git+https://github.com/davips/idict\n```\n\n### ...from source\n```bash\ngit clone https://github.com/davips/idict\ncd idict\npoetry install\npoetry install -E full  # use the flag \'full\' for extra functionality (recommended)\n```\n\n## Examples\n\n**Overview**\n<details>\n<p>\n\n```python3\n\n# Creation by direct instantiation.\nfrom idict import idict\n\nd = idict(x=5, y=7, z=10)\n\n# Creation from scratch.\n# The expression \'v >> a >> b\' means "Value \'v\' will be processed by step \'a\' then \'b\'".\n# A step can be a value insertion or a function application.\nfrom idict import empty\n\nd = empty >> {"x": 5} >> {"y": 7, "z": 10}\n\n# Empty alias (\'Ø\') usage.\nfrom idict import Ø\n\nd = Ø >> {"x": 5} >> {"y": 7, "z": 10}\nprint(d)\n"""\n{\n    "x": 5,\n    "y": 7,\n    "z": 10,\n    "id": "H8_d2866809b46b74333b8a1e77c2897466edc1b",\n    "ids": {\n        "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",\n        "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8",\n        "z": "ll_4dd7bf5575f1c410dc6458230cda99c380bda"\n    }\n}\n"""\n```\n\n```python3\n\n\n# Inverting color theme for a white background.\nfrom garoupa import setup\n\nsetup(dark_theme=False)\nd = idict(x=5, y=7, z=10)\nprint(d)\n\n\n"""\n{\n    "x": 5,\n    "y": 7,\n    "z": 10,\n    "id": "H8_d2866809b46b74333b8a1e77c2897466edc1b",\n    "ids": {\n        "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",\n        "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8",\n        "z": "ll_4dd7bf5575f1c410dc6458230cda99c380bda"\n    }\n}\n"""\n```\n\n```python3\n\n\n# Function application.\n# The input and output fields are detected from the function signature and returned dict.\ndef f(x):\n    return {"y": x ** 2}\n\n\nd2 = d >> f\nprint(d2)\n"""\n{\n    "y": "→(x)",\n    "x": 5,\n    "z": 10,\n    "id": "d6tEK5U0qq3sv0aCdSADNL3DiohLRgUAfdP7HEp4",\n    "ids": {\n        "y": "S8kpf19fdmWw65QIdKajRI2i03eLRgUAfdP7HEp4",\n        "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",\n        "z": "ll_4dd7bf5575f1c410dc6458230cda99c380bda"\n    }\n}\n"""\n```\n\n```python3\n\n\n# Anonymous function application.\nd2 = d >> (lambda y: {"y": y / 5})\nprint(d)\n"""\n{\n    "x": 5,\n    "y": 7,\n    "z": 10,\n    "id": "H8_d2866809b46b74333b8a1e77c2897466edc1b",\n    "ids": {\n        "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",\n        "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8",\n        "z": "ll_4dd7bf5575f1c410dc6458230cda99c380bda"\n    }\n}\n"""\n```\n\n```python3\n\n\n# Resulting values are evaluated lazily.\nd >>= lambda y: {"y": y / 5}\nprint(d.y)\n"""\n1.4\n"""\n```\n\n```python3\n\n\nprint(d)\n"""\n{\n    "y": 1.4,\n    "x": 5,\n    "z": 10,\n    "id": "OfV9fKxwEnz.wbUBTPj8HiLDPnT04cq1Fa6MPeo7",\n    "ids": {\n        "y": "LZHU.q4GvU4Y3YhITHVPLfKix2Q04cq1Fa6MPeo7",\n        "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",\n        "z": "ll_4dd7bf5575f1c410dc6458230cda99c380bda"\n    }\n}\n"""\n```\n\n```python3\n\n\n# Parameterized function application.\n# "Parameters" are distinguished from "fields" by having default values.\n# When the default value is None, it means it will be explicitly defined later by \'let\'.\nfrom idict import let\n\n\ndef f(x, y, a=None, b=None):\n    return {"z": a * x ** b, "w": y ** b}\n\n\nd2 = d >> let(f, a=7, b=2)\nprint(d2)\n"""\n{\n    "z": "→(a b x y)",\n    "w": "→(a b x y)",\n    "y": 1.4,\n    "x": 5,\n    "id": "0NVUmzD6.hrqvbHxGYHCrI8ZGKBXCQn7RgvPiD1z",\n    "ids": {\n        "z": "wlfii3B.l6NcROr6syPY3h-YOeRWyEZ5c6p3voFs",\n        "w": "ofEb.nRSYsUsgAnnyp4KYFovZaUOV6000sv....-",\n        "y": "LZHU.q4GvU4Y3YhITHVPLfKix2Q04cq1Fa6MPeo7",\n        "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977"\n    }\n}\n"""\n```\n\n```python3\n\n\n# Parameterized function application with sampling.\n# The default value is one of the following ranges, \n#     list, arithmetic progression, geometric progression.\n# Each parameter value will be sampled later.\n# A random number generator must be given.\nfrom idict import let\nfrom random import Random\n\n\ndef f(x, y, a=None, b=[1, 2, 3], ap=[1, 2, 3, ..., 10], gp=[1, 2, 4, ..., 16]):\n    return {"z": a * x ** b, "w": y ** ap * gp}\n\n\nd2 = d >> Random(0) >> let(f, a=7)\nprint(d2)\n"""\n{\n    "z": "→(a b ap gp x y)",\n    "w": "→(a b ap gp x y)",\n    "y": 1.4,\n    "x": 5,\n    "id": "sKzmHJm6wbwwLKYCewWyeMBQ8dAItpDRQXLcK1wt",\n    "ids": {\n        "z": "6hnZUsoJDO0UABeIu.OOWVu6p7vHpddQbNFsWO7n",\n        "w": "ofEb.nRSYsUsgAnnyp4KYFovZaUOV6000sv....-",\n        "y": "LZHU.q4GvU4Y3YhITHVPLfKix2Q04cq1Fa6MPeo7",\n        "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977"\n    }\n}\n"""\n```\n\n```python3\n\nprint(d2.z)\n"""\n175\n"""\n```\n\n```python3\n\nprint(d2)\n"""\n{\n    "z": 175,\n    "w": "10.541350399999995",\n    "y": 1.4,\n    "x": 5,\n    "id": "sKzmHJm6wbwwLKYCewWyeMBQ8dAItpDRQXLcK1wt",\n    "ids": {\n        "z": "6hnZUsoJDO0UABeIu.OOWVu6p7vHpddQbNFsWO7n",\n        "w": "ofEb.nRSYsUsgAnnyp4KYFovZaUOV6000sv....-",\n        "y": "LZHU.q4GvU4Y3YhITHVPLfKix2Q04cq1Fa6MPeo7",\n        "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977"\n    }\n}\n"""\n```\n\n\n</p>\n</details>\n\n**Identity example**\n<details>\n<p>\n\n```python3\nfrom idict import idict\n\na = idict(x=3)\nprint(a)\n"""\n{\n    "x": 3,\n    "id": "WB_e55a47230d67db81bcc1aecde8f1b950282cd",\n    "ids": {\n        "x": "WB_e55a47230d67db81bcc1aecde8f1b950282cd"\n    }\n}\n"""\n```\n\n```python3\n\nb = idict(y=5)\nprint(b)\n"""\n{\n    "y": 5,\n    "id": "0U_e2a86ff72e226d5365aea336044f7b4270977",\n    "ids": {\n        "y": "0U_e2a86ff72e226d5365aea336044f7b4270977"\n    }\n}\n"""\n```\n\n```python3\n\nprint(a >> b)\n"""\n{\n    "x": 3,\n    "y": 5,\n    "id": "Xt_a63010fa2b5b4c671270fbe8ec313568a8b35",\n    "ids": {\n        "x": "WB_e55a47230d67db81bcc1aecde8f1b950282cd",\n        "y": "0U_e2a86ff72e226d5365aea336044f7b4270977"\n    }\n}\n"""\n```\n\n\n</p>\n</details>\n\n**Merging two idicts**\n<details>\n<p>\n\n```python3\nfrom idict import idict\n\na = idict(x=3)\nprint(a)\n"""\n{\n    "x": 3,\n    "id": "WB_e55a47230d67db81bcc1aecde8f1b950282cd",\n    "ids": {\n        "x": "WB_e55a47230d67db81bcc1aecde8f1b950282cd"\n    }\n}\n"""\n```\n\n```python3\n\nb = idict(y=5)\nprint(b)\n"""\n{\n    "y": 5,\n    "id": "0U_e2a86ff72e226d5365aea336044f7b4270977",\n    "ids": {\n        "y": "0U_e2a86ff72e226d5365aea336044f7b4270977"\n    }\n}\n"""\n```\n\n```python3\n\nprint(a >> b)\n"""\n{\n    "x": 3,\n    "y": 5,\n    "id": "Xt_a63010fa2b5b4c671270fbe8ec313568a8b35",\n    "ids": {\n        "x": "WB_e55a47230d67db81bcc1aecde8f1b950282cd",\n        "y": "0U_e2a86ff72e226d5365aea336044f7b4270977"\n    }\n}\n"""\n```\n\n\n</p>\n</details>\n\n**Lazily applying functions to idict**\n<details>\n<p>\n\n```python3\nfrom idict import idict\n\na = idict(x=3)\nprint(a)\n"""\n{\n    "x": 3,\n    "id": "WB_e55a47230d67db81bcc1aecde8f1b950282cd",\n    "ids": {\n        "x": "WB_e55a47230d67db81bcc1aecde8f1b950282cd"\n    }\n}\n"""\n```\n\n```python3\n\na = a >> idict(y=5) >> {"z": 7} >> (lambda x, y, z: {"r": x ** y // z})\nprint(a)\n"""\n{\n    "r": "→(x y z)",\n    "x": 3,\n    "y": 5,\n    "z": 7,\n    "id": "H8DftZZ4nH6d67WSvYYxh-KsdBqp9MQBdvkLxU2o",\n    "ids": {\n        "r": "n57RGOgdv03kK4IqBkIf6oFrvgAp9MQBdvkLxU2o",\n        "x": "WB_e55a47230d67db81bcc1aecde8f1b950282cd",\n        "y": "0U_e2a86ff72e226d5365aea336044f7b4270977",\n        "z": "nX_da0e3a184cdeb1caf8778e34d26f5fd4cc8c8"\n    }\n}\n"""\n```\n\n```python3\n\nprint(a.r)\n"""\n34\n"""\n```\n\n```python3\n\nprint(a)\n"""\n{\n    "r": 34,\n    "x": 3,\n    "y": 5,\n    "z": 7,\n    "id": "H8DftZZ4nH6d67WSvYYxh-KsdBqp9MQBdvkLxU2o",\n    "ids": {\n        "r": "n57RGOgdv03kK4IqBkIf6oFrvgAp9MQBdvkLxU2o",\n        "x": "WB_e55a47230d67db81bcc1aecde8f1b950282cd",\n        "y": "0U_e2a86ff72e226d5365aea336044f7b4270977",\n        "z": "nX_da0e3a184cdeb1caf8778e34d26f5fd4cc8c8"\n    }\n}\n"""\n```\n\n\n</p>\n</details>\n\n**Parameterized functions and sampling**\n<details>\n<p>\n\n```python3\nfrom random import Random\n\nfrom idict import Ø, let\n\n\n# A function provide input fields and, optionally, parameters.\n# For instance:\n# \'a\' is sampled from an arithmetic progression\n# \'b\' is sampled from a geometric progression\n# Here, the syntax for default parameter values is borrowed with a new meaning.\ndef fun(x, y, a=[-100, -99, -98, ..., 100], b=[0.0001, 0.001, 0.01, ..., 100000000]):\n    return {"z": a * x + b * y}\n\n\ndef simplefun(x, y):\n    return {"z": x * y}\n\n\n# Creating an empty idict. Alternatively: d = idict().\nd = Ø >> {}\nd.show(colored=False)\n"""\n{\n    "id": "0000000000000000000000000000000000000000",\n    "ids": {}\n}\n"""\n```\n\n```python3\n\n# Putting some values. Alternatively: d = idict(x=5, y=7).\nd["x"] = 5\nd["y"] = 7\nd.show(colored=False)\n"""\n{\n    "x": 5,\n    "y": 7,\n    "id": "mP_2d615fd34f97ac906e162c6fc6aedadc4d140",\n    "ids": {\n        "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",\n        "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8"\n    }\n}\n"""\n```\n\n```python3\n\n# Parameter values are uniformly sampled.\nd1 = d >> simplefun\nd1.show(colored=False)\nprint(d1.z)\n"""\n{\n    "z": "→(x y)",\n    "x": 5,\n    "y": 7,\n    "id": "ZAasLu0lIEqhJyS1s8ML8WGeTnradBnjS7VNt6Mg",\n    "ids": {\n        "z": "iE6rHiYYwfwOBqa4Luh4XCd-myeadBnjS7VNt6Mg",\n        "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",\n        "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8"\n    }\n}\n35\n"""\n```\n\n```python3\n\nd2 = d >> simplefun\nd2.show(colored=False)\nprint(d2.z)\n"""\n{\n    "z": "→(x y)",\n    "x": 5,\n    "y": 7,\n    "id": "ZAasLu0lIEqhJyS1s8ML8WGeTnradBnjS7VNt6Mg",\n    "ids": {\n        "z": "iE6rHiYYwfwOBqa4Luh4XCd-myeadBnjS7VNt6Mg",\n        "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",\n        "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8"\n    }\n}\n35\n"""\n```\n\n```python3\n\n# Parameter values can also be manually set.\ne = d >> let(fun, a=5, b=10)\nprint(e.z)\n"""\n95\n"""\n```\n\n```python3\n\n# Not all parameters need to be set.\ne = d >> let(simplefun, a=5)\nprint(e.z)\n"""\n35\n"""\n```\n\n```python3\n\n# Each run will be a different sample for the missing parameters.\ne = e >> let(simplefun, a=5)\nprint(e.z)\n"""\n35\n"""\n```\n\n```python3\n\n# We can define the initial state of the random sampler.\n# It will be in effect from its location place onwards in the expression.\ne = d >> Random(0) >> let(fun, a=5)\nprint(e.z)\n"""\n725.0\n"""\n```\n\n```python3\n\n# All runs will yield the same result,\n# if starting from the same random number generator seed.\ne = e >> Random(0) >> let(fun, a=[555, 777])\nprint("Let \'a\' be a list:", e.z)\n"""\nLet \'a\' be a list: 700003885.0\n"""\n```\n\n```python3\n\n# Reproducible different runs are achievable by using a single random number generator.\ne = e >> Random(0) >> let(fun, a=[5, 25, 125, ..., 10000])\nprint("Let \'a\' be a geometric progression:", e.z)\n"""\nLet \'a\' be a geometric progression: 700003125.0\n"""\n```\n\n```python3\nrnd = Random(0)\ne = d >> rnd >> let(fun, a=5)\nprint(e.z)\ne = d >> rnd >> let(fun, a=5)  # Alternative syntax.\nprint(e.z)\n"""\n725.0\n700000025.0\n"""\n```\n\n\n</p>\n</details>\n\n**Composition of sets of functions**\n<details>\n<p>\n\n```python3\nfrom random import Random\n\nfrom idict import Ø\n\n\n# A multistep process can be defined without applying its functions\n\n\ndef g(x, y, a=[1, 2, 3, ..., 10], b=[0.00001, 0.0001, 0.001, ..., 100000]):\n    return {"z": a * x + b * y}\n\n\ndef h(z, c=[1, 2, 3]):\n    return {"z": c * z}\n\n\n# In the \'idict\' framework \'data is function\',\n# so the alias Ø represents the \'empty data object\' and the \'reflexive function\' at the same time.\n# In other words: \'inserting nothing\' has the same effect as \'doing nothing\'.\nfun = Ø >> g >> h  # \'empty\' or \'Ø\' enable the cartesian product of the subsequent sets of functions within the expression.\nprint(fun)\n"""\n«λ{} × λ»\n"""\n```\n\n```python3\n\n# Before a function is applied to a dict-like, the function free parameters remain unsampled.\n# The result is an ordered set of composite functions.\nd = {"x": 5, "y": 7} >> (Random(0) >> fun)\nprint(d)\n"""\n{\n    "z": "→(c z→(a b x y))",\n    "x": 5,\n    "y": 7,\n    "id": "9Cl8wQmnbvNBF8upBxrlH6Njosz0cq8CrzBYHeJO",\n    "ids": {\n        "z": "I3rJa4yiMmMpKq88VTYF1Kj3UCC0cq8CrzBYHeJO",\n        "x": ".T_f0bb8da3062cc75365ae0446044f7b3270977",\n        "y": "mX_dc5a686049ceb1caf8778e34d26f5fd4cc8c8"\n    }\n}\n"""\n```\n\n```python3\n\nprint(d.z)\n"""\n105.0\n"""\n```\n\n```python3\n\nd = {"x": 5, "y": 7} >> (Random(0) >> fun)\nprint(d.z)\n"""\n105.0\n"""\n```\n\n```python3\n\n# Reproducible different runs by passing a stateful random number generator.\nrnd = Random(0)\ne = d >> rnd >> fun\nprint(e.z)\n"""\n105.0\n"""\n```\n\n```python3\n\ne = d >> rnd >> fun\nprint(e.z)\n"""\n14050.0\n"""\n```\n\n```python3\n\n# Repeating the same results.\nrnd = Random(0)\ne = d >> rnd >> fun\nprint(e.z)\n"""\n105.0\n"""\n```\n\n```python3\n\ne = d >> rnd >> fun\nprint(e.z)\n"""\n14050.0\n"""\n```\n\n\n</p>\n</details>\n\n<persistence>\n\n## Concept\n\nAn `idict` is like a common Python `dict`, with extra functionality and lazy. \nIt is a mapping between string keys, called\nfields, and any serializable (pickable) object.\nEach `idict` has two extra entries: `id` (identifier) and `ids` (value identifiers).\n\nA custom 40-digit unique identifier (see [GaROUPa](https://pypi.org/project/garoupa))\ncan be provided as an attribute for each function.\nValue objects can have custom identifiers as well, if provided whithin the entry `ids`. \n\nOtherwise, identifiers for functions and values will be calculated through blake3 hashing of their content.\nFor functions, the bytecode is used as content. \nFor this reason, such functions should be simple, with minimal external dependencies or\nwith their import statements inside the function body.\nThis decreases the odds of using two functions with identical local code (and, therefore, identical identifiers) \nperforming different calculations.\n\n## Grants\n\nThis work was supported by Fapesp under supervision of\nProf. André C. P. L. F. de Carvalho at CEPID-CeMEAI (Grants 2013/07375-0 – 2019/01735-0)\nuntil 2021-03-31.\n',
    'author': 'davips',
    'author_email': 'dpsabc@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the idict project.
#  Please respect the license - more about this in the section (*) below.
#
#  idict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  idict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with idict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.
import pickle

import lz4.frame as lz4

from idict.config import GLOBAL


def pack(obj):
    r"""
    >>> from idict import setup
    >>> setup(compression_cachelimit_MB=0.000_100)
    >>> memo = GLOBAL["compression_cache"] = {}
    >>> GLOBAL["compression_cachesize"] = 0
    >>> b = b"000011"
    >>> pack(b)
    b'\x04"M\x18h@\x15\x00\x00\x00\x00\x00\x00\x006\x13\x00\x00\x00R\x80\x05\x95\n\x00\x01\x00\xa0C\x06000011\x94.\x00\x00\x00\x00'
    >>> memo[id(b)]["unpacked"]
    b'000011'
    >>> len(memo), GLOBAL["compression_cachesize"], GLOBAL["compression_cachelimit"]
    (1, 42, 100)
    >>> pack(b"asd")
    b'\x04"M\x18h@\x12\x00\x00\x00\x00\x00\x00\x00\xd9\x10\x00\x00\x00R\x80\x05\x95\x07\x00\x01\x00pC\x03asd\x94.\x00\x00\x00\x00'
    >>> len(memo), GLOBAL["compression_cachesize"], GLOBAL["compression_cachelimit"]
    (2, 81, 100)
    >>> len(pack(b"123"))
    39
    >>> len(memo), GLOBAL["compression_cachesize"], GLOBAL["compression_cachelimit"]
    (2, 78, 100)
    """
    memid = id(obj)
    memo = GLOBAL["compression_cache"]
    if memid in memo:
        if obj is memo[memid]["unpacked"]:
            return memo[memid]["packed"]
        else:
            # rare collision
            GLOBAL["compression_cachesize"] -= memo[memid]["packed"]
            del memo[memid]

    try:
        dump = pickle.dumps(obj, protocol=5)
        blob = lz4.compress(dump)
        GLOBAL["compression_cachesize"] += len(blob)
        memo[memid] = {"unpacked": obj, "packed": blob}

        # LRU
        keys = iter(list(memo.keys()))
        while len(memo) > 0 and GLOBAL["compression_cachesize"] > GLOBAL["compression_cachelimit"]:
            k = next(keys)
            v = memo.pop(k)["packed"]
            GLOBAL["compression_cachesize"] -= len(v)

        return blob
    except KeyError as e:  # pragma: no cover
        if str(e) == "'__getstate__'":
            raise Exception("Unpickable value:", type(obj))
        else:
            raise e


def unpack(blob):
    r"""
    >>> unpack(b'\x04"M\x18h@\x15\x00\x00\x00\x00\x00\x00\x006\x13\x00\x00\x00R\x80\x05\x95\n\x00\x01\x00\xa0C\x06000011\x94.\x00\x00\x00\x00')
    b'000011'
    """
    return pickle.loads(lz4.decompress(blob))
